; DivError.asm
; For locking mathematical errors in Borland Pascal
; (C) Denisov A.W. 15.06.1998
; Version 1.00

public div_init
public AssignMathError,UnassignMathError

; ======== CONSTANTS =============================

EOI                            equ 20h
first_controller               equ 20h
second_controller              equ 0A0h

; ================================================

assume cs:code,ds:data

data segment word public

  EXTRN ExitProc

data ends


LOCALS
code segment byte public

; ========================================================
; === Data in code segment / for exlude of segment of data

  f_handlers_ptr  dd 6 dup(0) ; if assigned, call the program(s)
  i_handler_ptr   dd 0

  math_errors     dw 204,500,200,205,206,501
  FPU_environment dw ?
  in_fpu_isr      dw 0

  old_00h         dd ?  ; "Divide Overflow" interrupt number
  old_75h         dd ?  ; "FPU error" interrupt number
  OldExitProc     dd ?  ; Old Exit Procedure

;=========================================================

;==================== div286_fault_isr / INTERRUPT / ================
div286_fault_isr proc

  sti
  push  bx
  push  cx
  push  dx
  push  si
  push  di
  push  ds
  push  es

  mov   bx,sp
  add   bx,14
  mov   di,ss:[bx]
  mov   es,ss:[bx + 2]
  lds   si,ss:[bx]
  cld
  lodsw
  and   ah,0C0h
  cmp   ah,0C0h
  je    @@10
  add   word ptr ss:[bx],2
 @@10:
  add   word ptr ss:[bx],2

  mov   ax,word ptr cs:i_handler_ptr
  mov   bx,word ptr cs:i_handler_ptr + 2
  or    ax,bx
  jz    @@Exit

  mov   ax,200
  push  ax
  mov   ax,data
  mov   ds,ax
  call  [i_handler_ptr]

 @@Exit:
  pop   es
  pop   ds
  pop   di
  pop   si
  pop   dx
  pop   cx
  pop   bx
  xor   ax,ax

  iret

div286_fault_isr endp

;================= float_error_isr / INTERRUPT / ====================
float_error_isr proc

  sti
  cmp     cs:in_fpu_isr,0
  je      @@10
  iret

 @@10:
  mov     cs:in_fpu_isr,1
  mov     al,EOI
  out     first_controller,al
  jmp     $ + 2
  out     second_controller,al

  fnstsw  cs:FPU_environment
  fnclex          ; FPU reset
  fwait           ; FPU syncronize

  push    bx
  push    cx
  push    dx
  push    si
  push    di
  push    ds
  push    es

  mov     ax,cs:FPU_environment
  and     al,00111111b

  mov     bx,offset cs:f_handlers_ptr
  mov     di,offset math_errors
  mov     cx,6
 @@calling:
  test    al,1
  jz      @@20

  mov     ax,cs:[bx]
  mov     si,cs:[bx + 2]
  or      ax,si
  jz      @@Exit
  mov     ax,cs:[di]
  push    ax
  mov     ax,data
  mov     ds,ax
  call    dword ptr cs:[bx]
  jmp short @@Exit
 @@20:
  shr     al,1
  add     bx,4
  add     di,2
  loop  @@calling

 @@Exit:
  pop     es
  pop     ds
  pop     di
  pop     si
  pop     dx
  pop     cx
  pop     bx

  xor     ax,ax

  cli
  mov     cs:in_fpu_isr,0

  iret

float_error_isr endp

;===================== div_init / NEAR / ============================
div_init proc near

  mov   ax,word ptr ExitProc
  mov   bx,word ptr ExitProc + 2

  mov   word ptr ExitProc + 2,cs
  mov   word ptr ExitProc,offset cs:div_uninstall

  push  cs
  pop   ds

  mov   word ptr OldExitProc,ax
  mov   word ptr OldExitProc + 2,bx

  mov   ax,3500h
  int   21h
  mov   word ptr old_00h,bx
  mov   word ptr old_00h + 2,es
  mov   ax,3575h
  int   21h
  mov   word ptr old_75h,bx
  mov   word ptr old_75h + 2,es

  mov   dx,offset div286_fault_isr
  mov   ax,2500h
  int   21h
  mov   dx,offset float_error_isr
  mov   ax,2575h
  int   21h

  mov   ax,data
  mov   ds,ax

  retn

div_init endp

;================== div_uninstall / FAR / ===========================
div_uninstall proc far

  lds   dx,old_00h
  mov   ax,2500h
  int   21h
  lds   dx,old_75h
  mov   ax,2575h
  int   21h

  mov   ax,data
  mov   ds,ax

  mov   ax,word ptr cs:OldExitProc
  mov   bx,word ptr cs:OldExitProc + 2

  mov   word ptr ExitProc,ax
  mov   word ptr ExitProc + 2,bx

  retf

div_uninstall endp

;=============== AssignMathError / FAR / ============================
AssignMathError proc far

  mov   bx,sp
  mov   al,ss:[bx + 8]
  mov   cl,4
  mul   cl
  les   di,ss:[bx + 4]
  mov   bx,offset cs:f_handlers_ptr
  add   bx,ax
  mov   cs:[bx],di
  mov   cs:[bx + 2],es

  retf  6

AssignMathError endp

;=============== UnassignMathError / FAR / ==========================
UnassignMathError proc far

  mov   bx,sp
  mov   cx,ss:[bx + 4]
  mov   al,4
  mul   cl

  mov   bx,offset cs:f_handlers_ptr
  add   bx,ax

  xor   ax,ax
  mov   cs:[bx],ax
  mov   cs:[bx + 2],ax

  retf  2

UnassignMathError endp


code ends

end